package ru.dbotthepony.mc.otm.container

import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.core.collect.iterator
import ru.dbotthepony.mc.otm.core.collect.nonEmpty

operator fun Container.set(index: Int, value: ItemStack) = setItem(index, value)
operator fun Container.get(index: Int): ItemStack = getItem(index)

fun Container.addItem(stack: ItemStack, range: IntRange, simulate: Boolean = false): ItemStack {
	if (this is MatteryContainer) {
		return this.addItem(stack, range, simulate)
	}

	if (range.last >= containerSize || range.first < 0)
		throw IllegalArgumentException("Invalid range: $range")

	if (stack.isEmpty)
		return stack

	val copy = stack.copy()

	// двигаем в одинаковые слоты
	for (slot in range) {
		if (ItemStack.isSameItemSameTags(this[slot], copy)) {
			val slotStack = this[slot]
			val slotLimit = maxStackSize.coerceAtMost(slotStack.maxStackSize)

			if (slotStack.count < slotLimit) {
				val newCount = (slotStack.count + copy.count).coerceAtMost(slotLimit)
				val diff = newCount - slotStack.count

				if (!simulate) {
					slotStack.count = newCount
					setChanged()
				}

				copy.shrink(diff)

				if (copy.isEmpty) {
					return copy
				}
			}
		}
	}

	// двигаем в пустые слоты
	for (slot in range) {
		if (this[slot].isEmpty) {
			val diff = copy.count.coerceAtMost(maxStackSize.coerceAtMost(copy.maxStackSize))

			if (!simulate) {
				val copyToPut = copy.copy()
				copyToPut.count = diff
				this[slot] = copyToPut
				setChanged()
			}

			copy.shrink(diff)

			if (copy.isEmpty) {
				return copy
			}
		}
	}

	return copy
}

fun Container.addItem(stack: ItemStack, simulate: Boolean): ItemStack = addItem(stack, 0 until containerSize, simulate)

inline fun <T> Container.forEach(cap: Capability<T>, consumer: (Pair<ItemStack, T>) -> Unit) {
	for (value in iterator(cap)) {
		consumer(value)
	}
}

inline fun <T> Container.forEach(cap: Capability<T>, consumer: (ItemStack, T) -> Unit) {
	for ((a, b) in iterator(cap)) {
		consumer(a, b)
	}
}

inline fun <T> Container.forEachItem(cap: Capability<T>, consumer: (ItemStack) -> Unit) {
	for (pair in iterator(cap)) {
		consumer(pair.first)
	}
}

inline fun <T> Container.forEachCapability(cap: Capability<T>, consumer: (T) -> Unit) {
	for (pair in iterator(cap)) {
		consumer(pair.second)
	}
}

inline fun Container.forEach(lambda: (ItemStack) -> Unit) {
	for (value in iterator()) {
		lambda(value)
	}
}

inline fun Container.forEachNonEmpty(lambda: (ItemStack) -> Unit) {
	for (value in iterator().nonEmpty()) {
		lambda(value)
	}
}
